#ifndef _xobject_h_
#define _xobject_h_

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <io.h>
#include <math.h>

#define IS_DIGIT(a) ((a)>='0' && (a)<='9')
#define IS_ALPHA(a) (((a)>='A' && (a)<='Z') || ((a)>='a' && (a)<='z') || (a)=='_')
#define IS_AD(a) (IS_ALPHA(a) || IS_DIGIT(a))

#define NAME_LENGTH	32
#define TOKEN_SEPARATOR	"|"

#define CLASS_VOID		0
#define CLASS_INT		1
#define CLASS_DOUBLE		2
#define CLASS_STRING		3
#define CLASS_GROUP		4
#define CLASS_NAMEDGROUP	5

#define FLAGS_NORMAL		0x0
#define FLAGS_TRANSIENT		0x1
#define FLAGS_READONLY		0x2
#define FLAGS_UPDATEONLY	0x4
#define FLAGS_TEXT		0x8
#define FLAGS_BUSY		0x10

#define CLEAN_PRESERVE_NONE	0
#define CLEAN_PRESERVE_ALL	-1
#define CLEAN_PRESERVE_EOL	0x1

#define LOG_INFO	3
#define LOG_MSG		2
#define LOG_ERROR	1
#define LOG_NONE	0
#define LOG_ALL		-1

#define CLASS_VOID_NAME		"void"
#define CLASS_INT_NAME		"int"
#define CLASS_DOUBLE_NAME	"double"
#define CLASS_STRING_NAME	"string"
#define CLASS_GROUP_NAME	"group"
#define CLASS_NAMEDGROUP_NAME	"namedgroup"

#define OBJECT_VERSION		"version"

#define NAMEDGROUP_TITLE	"title"
#define NAMEDGROUP_TYPE		"type"
#define NAMEDGROUP_ELEMENTS	"elements"

struct PARSEINFO {
	char*	data;
	int	size;
	int	ptr;
	int	line;
	int	pos;
	int	level;
	int	newclass;
	int	flags;
};

class _Export XObject {
public:
	static void setLogLevel(int);
	static int log(char*, ...);
	static int log(int, char*, ...);
	static int logObject(XObject*, char* = NULL);
	static void logClasses();

	static void	initClass();
	static void	unInitClass();
	static int	loadFile(char*, char**, int*);
	static int	saveFile(char*, char*, int);
	static XObject*	getInstance(char*, char*, int = FLAGS_NORMAL, int =-1);
	static XObject* getInstanceText(char*, int*, int =-1);
	static XObject* getInstanceBin(char*, int*, int = -1);
	static int	loadClasses(char*);
	static XObject* loadObject(char*);
	static int	saveObject(char*, XObject*);
private:
	static int logLevel;
	static int initCounter;
	static int objectCounter;
	static int classesNum;
	static XObject** classes;
	static int escSeqsNum;
	static char escSeqs[];
	static int kwdFlagsNum;
	static char* kwdFlags[];
public:
	static int	getClassNum();
	static char*	getClassName(int);
protected:
	static XObject* txtParse(PARSEINFO*, XObject*);
	static int txtSkipSpaces(PARSEINFO*);
	static int txtReadHead(PARSEINFO*, char*, char*, int*, int*);

	static XObject* binParse(PARSEINFO*, XObject*);
	static int binReadHead(PARSEINFO*, char*, char*, int*, int*);

	static XObject* getClass(int);
	static XObject* getClass(char*);
	static XObject* getClass(char*, int);
	static int	addClass(XObject*, char *, int);

protected:
	char name[NAME_LENGTH];
	int classID;
	int flags;	//transient

	union {
		struct { double lo; } dval;
		struct { int lo; int hi; } ival;
		struct { int size; char* str; } sval;
		struct { int size; XObject** group; } gval;
	};

protected:
	XObject* getClass();

	XObject();

public:
	virtual	~XObject();
	int	isVoid();
	int	isInteger();
	int	isDouble();
	int	isString();
	int	isComplex();
	int	isGroup();
	int	isObject();
	int	isObject(char*);
	int	isBasic();

	virtual XObject*	clone();
	int	initFrom(XObject*);

	char*	getName();
	int	setName(char* n);
	char*	getClassName();
	int	getClassVersion();

	int	getFlags();
	int	setFlags(int);
	void	flagSet(int);
	void	flagClear(int);
	int	flagIsSet(int);

	int	getSizeBin();
	int	getSizeText();

	int	getIntValue();
	int	setIntValue(int);

	double	getDoubleValue();
	int	setDoubleValue(double);

	char*	getStringValue();
	int	getStringSize();
	int	getStringLength();
	int	setStringSize(int);
	int	setStringValue(int, char*);
	int	setStringValue(char*);

	int	getGroupSize();
	XObject*	getGroupElement(int);
	XObject*	getGroupElement(char*);
	XObject*	getGroupElement2(char*);
	int	addGroupElement(XObject*);
	int 	removeGroupElement(int);
	int 	removeGroup();

	int	loadBinFile(char*);
	int	readText(char*, int);
	int	readBin(char*, int);
	int	writeBin(char*, int);
	int	writeText(char*, int, int = 0);

	virtual XObject* getValue();
	virtual int show(int);
	virtual int print(int);

// misc
	static char*	strPrintInteger(char*, int, int);
	static char*	strPrintDouble(char*, double, int, int);
	static char*	strPrintString(char*, char*, int, int*);
	static char*	strPrintStringEsc(char*, char*, int, int*);
protected:
	int	clear();
	int 	removeGroup2();
	int	copyFrom(XObject*);

protected:
	int	getSizeBin2();
	int	writeBin2(char*, int);
};



#endif